'=======================
' VPinMAME driver code.
'=======================
Option Explicit
Const VPinMAMEDriverVer = 0.997
Dim Controller   ' VPinMAME Controller Object
Dim vpmTimer     ' Timer Object
Dim vpmNudge     ' Nudge handler Object
Dim Lights(88)   ' Put all lamps in an array for easier handling

' Callbacks
Dim SolCallback(64) ' Solenoids (parsed at Runtime)
Dim LampCallback    ' Called after lamps are updated
Dim GICallback      ' Called for each changed GI String
Dim MotorCallback   ' Called after solenoids are updated
' Other Game specific info
Dim ExtraKeyHelp    ' Help string for game specific keys

'-----------------------------------------------------------------------------
' These helper function requires the following objects on the table:
'   PinMAMETimer : Timer object
'   PulseTimer   : Timer object
'
'
' Available classes:
' ------------------
' cvpmTimer (Object = vpmTimer)
'   (Public)  .PulseSwitch   - pulse switch and call callback after delay (default)
'   (Public)  .AddTimer      - call callback after delay
'   (Public)  .Reset         - Re-set all ballStacks
'   (Friend)  .AddBallStack  - add a vpmBallStack class (from vpmBallStack class)
'   (Private) .Update        - called from timer event
'
' cvpmBallStack (Create as many as needed)
'   (Public) .InitSw        - init switches used in stack
'   (Public) .InitSaucer    - init saucer
'   (Public) .InitKick      - init exit kicker
'   (Public) .KickBalls     - Maximum number of balls kicked out at the same time
'   (Public) .KickForceVar  - Initial ExitKicker Force value varies by this much (+/-, minimum force = 1)
'   (Public) .KickAngleVar  - ExitKicker Angle value varies by this much (+/-)
'   (Public) .InitEntrySnd  - Sounds for Entry kicker
'   (Public) .InitExitSnd   - Sounds for Entry kicker
'   (Public) .AddBall       - add ball in "kicker" to stack
'   (Public) .SolIn         - Solenoid handler for entry solenoid
'   (Public) .EntrySol_On   - entry solenoid fired
'   (Public) .SolOut        - Solenoid handler for exit solenoid
'   (Public) .ExitSol_On    - exit solenoid fired
'   (Public) .Balls         - get/set number of balls in stack (default)
'   (Public) .BallsPending  - get number of balls waiting to come in to stack
'   (Public) .Reset         - reset and update all ballstack switches
'   (Friend) .Update        - Update ball positions (from vpmTimer class)
'   (Public) .SolExit       - exit solenoid handler
'   (Public) .SolEntry      - Entry solenoid handler

' cvpmNudge (Object = vpmNudge)
'   Hopefully we can add a real pendulum simulator in the future
'   (Public)  .TiltSwitch = x     - set tilt switch
'   (Public)  .Senitivity = x     - Set tiltsensitivity (0-10)
'   (Public)  .DoNudge dir,power  - Nudge table
'   (Private) .TimeOut            - Handle tilting
'
' cvpmDropTarget (create as many as needed)
'   (Public)  .InitDrop     - initialise DropTarget bank
'   (Public)  .InitSnd      - sound to use for targets
'   (Public)  .Hit          - A target had been hit
'   (Public)  .SolDropUp    - Solenoid handler for Bank reset
'   (Public)  .DropSol_On   - Reset target bank
'
' Generic solenoid handlers:
' --------------------------
' vpmSolFlipper flipObj1, flipObj2  - "flips flippers". Set unused to Nothing
' vpmSolDiverter divObj, sound      - open/close diverter (flipper) with/without sound
' vpmSolWall wallObj, sound         - Raise/Drop wall with/without sound
' vpmSolAutoPlunger plungerObj      - Autoplunger/kickback
' vpmSolSound sound                 - Play sound only
' vpmFlasher flashObj               - Flashes flasher
' vpmSolMagnet magnetObj            - Enables kicker object/kick ball down
'
' Variables declared (to be filled in):
' ---------------------------------------
' SolCallback()  - handler for each solenoid
' Lights()       - Lamps
'
' Constants used (must be defined):
' ---------------------------------
' UseSync           - Stop/Start PinMAME while updating table
' UseLamps          - Update lamps
' LampCallback      - Sub to call after lamps are updated
'                     (To update multiple lamps connected to same wire)
' GICallback        - Sub to call to update GI strings
' MotorCallback     - Called once every update
' SFlipperOn        - Flipper activate sound
' SFlipperOff       - Flipper deactivate sound
' SSolenoidOn       - Solenoid activate sound
' SSolenoidOff      - Solenoid deactivate sound
' SCoin             - Coin Sound
'
' Exported variables:
' -------------------
' vpmTimer      - Timer class for PulsSwitch etc
' vpmNudge      - Class for table nudge handling
'
' Exported constants:
' -------------------
' WPC/S11 keys
' WPC/S11 cabinet switches
' Flipper solenoids
'-----------------------------------------------------
'
' Revision History
' 24-Mar-01  First working version
' 26-Mar-01  Changed Timer callback to allow extra arguments (like solenoids)
' 28-Mar-01  Changed timer handling to use a separate timer.
'  7-Apr-01  Added Kickballs and BallStack now allows no exit kicker
'  9-Apr-01  (Dorsola) Added KickForceVar and KickAngleVar to specify variances in
'            force and angle for Ballstack exit kicker.
' 11-Apr-01  Added version variable, corrected multiple balls kickout
' 16-Apr-01  Added Reset method to restore switches after a reset
'            Added MotorCallback
' 19-Apr-01  Added SolEntry, SolExit methods to ballstack
'            Added vpmSolAutoPlunger and Saucer handling
' 22-Apr-01  Bugfixing and cSaucer now disables ExitKicker when ball lands
'            (To prevent multiple balls in kicker)
'            Prefixed all solenoid handlers with "vpm"
'  3-May-01  Wall and flasher now accepts arrays
'---------------------------------------------------------------------------------
Private Const PinMameInterval = 1 ' Light/solnoids run as quick as possible

Private Const cMaxStacks  = 8
Private Const cSwPulse    = 1
Private Const cMaxStackSw = 7
Private Const cMaxBalls   = 10
Private Const cMaxTimers  = 20 ' Spinners can generate a lot of timers
Private Const cTimerPulse = 40 ' Timer runs at 25Hz

Class cvpmTimer
	Private cTime(), cDelay(), cSw(), cCallback(), cNow, cTimers, cBallStack()

	Private Sub Class_Initialize
		ReDim cTime(cMaxTimers), cDelay(cMaxTimers), cSw(cMaxTimers), cCallback(cMaxTimers), cBallStack(cMaxTimers)
		cNow = 0 : cTimers = 0
		PulseTimer.Interval = cTimerPulse : PulseTimer.Enabled = True
	End Sub

	Private Sub DeleteTimer(pos)
		Dim ii
		cTimers = cTimers - 1
		For ii = pos To cTimers
			cTime(ii) = cTime(ii+1) : cSw(ii) = cSw(ii+1) : cCallback(ii) = cCallback(ii+1) : cDelay(ii) = cDelay(ii+1)
		Next
	End sub

	Public Sub Reset
		Dim ii
		For ii = 1 To cMaxStacks
			If isObject(cBallStack(ii)) Then cBallStack(ii).Reset : Else Exit For
		Next
	End Sub

	Public Sub Update
		Dim ii, jj, sw

		For ii = 1 To cMaxStacks
			If isObject(cBallStack(ii)) Then cBallStack(ii).Update() Else Exit For
		Next

		If cTimers Then cNow = cNow + 1 Else Exit Sub
		For ii = 1 To cTimers ' Set pending switches
			If cSw(ii) < 0 Then cSw(ii) = -cSw(ii) : Controller.Switch(cSw(ii)) = True : cTime(ii) = cNow + cSwPulse
		Next
		ii = 1
		Do While ii <= cTimers
			If cTime(ii) = cNow Then ' Timer has expired
				sw = cSw(ii)
				If cDelay(ii) Then
					If sw > 0 Then Controller.Switch(sw) = False : cTime(ii) = cNow + cDelay(ii) : cDelay(ii) = 0
				Else
					If varType(cCallback(ii)) = vbString And cCallback(ii) > "" Then Execute cCallback(ii) & " " & sw & " "
					DeleteTimer ii : ii = ii - 1
				End If
				If sw > 0 Then ' Check if there are more timers for this switch
					For jj = ii+1 To cTimers
						If sw = cSw(jj) And cDelay(jj) > 0 Then cSw(jj) = -cSw(jj) : Exit For
					Next
				End If
			End If
			ii = ii + 1
		Loop
	End Sub

	Public Sub AddBallStack(stack)
		Dim ii
		For ii = 1 To cMaxStacks
			If Not isObject(cBallStack(ii)) Then Set cBallStack(ii) = stack : Exit For
		Next
	End Sub

	Public Default Sub PulseSwitch(swNo, delay, callback)
		Dim ii
		if cTimers = cMaxTimers Then Exit Sub
		cTimers = cTimers + 1
		cSw(cTimers) = -swNo : cCallback(cTimers) = callback : cTime(cTimers) = cNow + cSwPulse
		If delay = 0 Then cDelay(cTimers) = 1 Else cDelay(cTimers) = delay \ cTimerPulse
		' Check if event already present for switch
		For ii = 1 To cTimers-1
			If (cSw(ii) = swNo Or cSw(ii) = -swNo) And cDelay(ii) > 0 Then
				cTime(cTimers) = 0 : cSw(cTimers) = swNo : Exit For
			End If
		Next
	End Sub

	Public Sub AddTimer(delay, callback)
		If cTimers = cMaxTimers Then Exit Sub
		cTimers = cTimers + 1
		cSw(cTimers) = 0 : cCallback(cTimers) = callback : cDelay(cTimers) = 0
		If delay \ cTimerPulse = 0 Then delay = cTimerPulse:
		cTime(cTimers) = cNow + delay \ cTimerPulse :
	End Sub
End Class

Class cvpmBallStack
	Private cSw(), cProxySw, cProxyNo, cEntrySw
	Private cExitKicker, cExitKickDir, cExitKickPower, cKickBalls
	Private cKickForceVar, cKickAngleVar
	Private cBalls, cBallIn
	Private cBallPos()
	Private cUpdate, cNow, cSaucer
	Private cEntrySnd, cEntrySndBall, cExitSnd, cExitSndBall

	Private Sub Class_Initialize
		ReDim cSw(cMaxStackSw), cBallPos(cMaxBalls)
		cBallIn = 0 : cBalls = 0 : cUpdate = False : cProxyNo = 0
		Set cExitKicker = Nothing : cKickBalls = 1 : cSaucer = False
		cKickForceVar = 0 : cKickAngleVar = 0
		cEntrySnd = 0 : cEntrySndBall = 0 : cExitSnd = 0 : cExitSndBall = 0
		vpmTimer.AddBallStack Me
	End Sub

	Private Function SetSw(no, status)
		SetSw = False
		If no <= cMaxStackSw Then If cSw(no) Then Controller.Switch(cSw(no)) = status : SetSw = True
	End Function

	Private Function SetProxy(pos, status)
		SetProxy = False
		If pos = cProxyNo Then Controller.Switch(cProxySw) = status : SetProxy = True
	End Function

	Public Sub Reset
		Dim ii
		For ii = 1 to cBalls : SetSw cBallPos(ii), True : SetProxy cBallPos(ii), True : Next
	End Sub

	Public Sub Update
		Dim BallQue, ii

		If cUpdate Then cUpdate = False Else Exit Sub
		BallQue = 1
		For ii = 1 To cBalls
			If cBallpos(ii) > BallQue Then
				SetSw cBallPos(ii), False : SetProxy cBallPos(ii), False
				Do
					cBallPos(ii) = cBallPos(ii) - 1
				Loop Until SetSw(cBallPos(ii), True) Or SetProxy(cBallPos(ii), True) Or cBallPos(ii) = BallQue
				cUpdate = True
			End If
			BallQue = cBallPos(ii) + 1
		Next
	End Sub

	Public Sub AddBall(kicker)
		If cSaucer Then cExitKicker.Enabled = False Else If isObject(kicker) Then kicker.DestroyBall
		If cEntrySw Then
			Controller.Switch(cEntrySw) = True : cBallIn = cBallIn + 1
		Else
			cBalls = cBalls + 1 : cBallPos(cBalls) = cMaxStackSw + 1 : cUpdate = True
		End If
	End Sub

	Public Sub SolIn(enabled)
		If enabled Then
			If cBallIn Then PlaySound cEntrySndBall Else PlaySound cEntrySnd
			EntrySol_On
		End If
	End Sub

	Public Sub EntrySol_On
		If cBallIn = 0 Then Exit Sub
		cBalls = cBalls + 1 : cBallIn = cBallIn - 1 : cBallPos(cBalls) = cMaxStackSw + 1 : cUpdate = True
		If cEntrySw And cBallIn = 0 Then Controller.Switch(cEntrySw) = False
	End Sub

	Public Sub SolOut(enabled)
		If enabled Then
			If cBalls Then PlaySound cExitSndBall Else PlaySound cExitSnd
			ExitSol_On
		End If
	End Sub

	Public Sub ExitSol_On
		Dim ii,jj
		Dim kForce, kAngle

		If cBalls = 0 Then Exit Sub

		kForce = cExitKickPower + (Rnd - 0.5)*cKickForceVar
		If cSaucer Then
			SetProxy 1, False : SetSw 1, False
			cBalls = 0 : kAngle = cExitKickDir + (Rnd - 0.5)*cKickAngleVar
			cExitKicker.Kick kAngle, kForce
			cExitKicker.Enabled = True
		Else
			For ii = 1 To cKickBalls
				If cBalls = 0 Or cBallPos(1) <> ii Then Exit For ' No more balls
				For jj = 2 To cBalls	' Move balls in array
					cBallPos(jj-1) = cBallPos(jj)
				Next
				cBallPos(cBalls) = 0 : cBalls = cBalls - 1 : cUpdate = True
				SetProxy ii, False : SetSw ii, False
				If Not cExitKicker Is Nothing Then
					If kForce < 1 Then kForce = 1
					kAngle = cExitKickDir + (Rnd - 0.5)*cKickAngleVar
					vpmTimer.AddTimer (ii-1)*200, cExitKicker.Name & ".CreateBall : " &_
					cExitKicker.Name & ".Kick " & CInt(kAngle) & "," & CInt(kForce+0.49) & "+"
				End If
				kForce = KForce * 0.6
			Next
		End If
	End Sub

	Public Sub InitSaucer(kicker, sw, dir, power)
		cSw(1) = sw : InitKick kicker, dir, power : cSaucer = True
	End Sub

	Public Sub InitSw(entry, sw1, sw2, sw3, sw4, sw5, sw6, sw7)
		cEntrySw = Entry : cSw(1) = Sw1 : cSw(2) = Sw2 : cSw(3) = Sw3 : cSw(4) = Sw4
		cSw(5) = Sw5 : cSw(6) = Sw6 : cSw(7) = Sw7
	End Sub

	Public Sub InitKick(kicker, dir, power)
		Set cExitKicker = kicker : cExitKickDir = dir : cExitKickPower = power
	End Sub

	Public Sub InitProxy(proxyPos, swNo)
		cProxySw = swNo
		cProxyNo = proxyPos
	End Sub

    Public Sub InitEntrySnd(ball, noBall)
	    cEntrySndBall = ball : cEntrySnd = noBall
    End Sub

    Public Sub InitExitSnd(ball, noBall)
	    cExitSndBall = ball : cExitSnd = noBall
    End Sub

	Public Property Let Balls(noballs)
		Dim ii
		For ii = 1 To cMaxStackSw
			If cSw(ii) Then Controller.Switch(cSw(ii)) = False
			cBallPos(ii) = 8
		Next
		If cProxyNo Then Controller.Switch(cProxySw) = False
		cBalls = noBalls : cUpdate = True
	End Property

	Public Default Property Get Balls
		Balls = cBalls
	End Property

	Public Property Get BallsPending
		BallsPending = cBallIn
	End Property

	Public Property Let KickBalls(noBalls)
		cKickBalls = noBalls
	End Property

	Public Property Let KickForceVar(varAmt)
		cKickForceVar = varAmt
	End Property

	Public Property Let KickAngleVar(varAmt)
		cKickAngleVar = varAmt
	End Property

    ' Obsolete stuff
	Public Sub SolEntry(sound1, sound2, enabled)
		If enabled Then
			If cBallIn Then PlaySound sound1 Else PlaySound sound2
			EntrySol_On
		End If
	End Sub
	Public Sub SolExit(sound1, sound2, enabled)
		If enabled Then
			If cBallIn Then PlaySound sound1 Else PlaySound sound2
			ExitSol_On
		End If
	End Sub
End Class

class cvpmNudge
	Dim cCount, cSensitivity, cTiltSwNo

	Private Sub Class_Initialize
		cCount = 0 : cTiltSwNo = 0 : cSensitivity = 3
	End sub

	Property Let TiltSwitch(swNo)
		cTiltSwNo = swNo
	End property

	Property Let Sensitivity(sens)
		cSensitivity = (11-sens)\2+1
	End property

	Public Sub DoNudge(dir, power)
		Nudge dir, power
		If cTiltSwNo = 0 Then Exit Sub ' If no switch why care
		cCount = cCount + 1
		If cCount >= cSensitivity Then
			vpmTimer.PulseSwitch cTiltSwNo, 3000, "vpmNudge.TimeOut"
		Else
			vpmTimer.AddTimer 3000, "vpmNudge.TimeOut"
		End If
	End sub

	Public Sub TimeOut(sw)
		If cCount Then cCount = cCount - 1
	End Sub
End Class

Class cvpmDropTarget
	Dim cDropObj, cDropSw, cDropSnd, cRaiseSnd
	Public Sub InitDrop(drops,sw)
		cDropObj = drops : cDropSw = sw : cDropSnd = 0 : cRaiseSnd = 0
	End Sub

	Public Sub InitSnd(drop, raise)
		cDropSnd = drop : cRaiseSnd = raise
	End Sub

	Public Sub Hit(no)
		PlaySound cDropSnd : cDropObj(no-1).IsDropped = True : Controller.Switch(cDropSw(no-1)) = True
	End Sub

	Public Sub SolDropUp(enabled)
		Dim ii
                If Enabled Then
                        PlaySound cRaiseSnd
		        For Each ii In cDropObj : ii.IsDropped = False : Next
		        For Each ii In cDropSw  : Controller.Switch(ii) = False : Next
                End If
	End Sub

    Public Sub DropSol_On
        SolDropUp(True)
    End Sub
End Class

Sub PulseTimer_Timer()
	vpmTimer.Update
End Sub
' End of classes

'-------------------
' Main Loop
'------------------
Private Const LAMPNO = 0
Private Const LAMPSTATE = 1
Private Const SOLNO = 0
Private Const SOLSTATE = 1
Private Const GISTRINGNO = 0
Private Const GISTRINGSTATE = 1

Sub PinMAMETimer_Timer()
	Dim Changed, Count, ii, funcName
	Me.Enabled = False
	If UseSync Then Controller.Pause = True
	If UseLamps Then
		Changed = Controller.ChangedLamps
		If Not IsEmpty(Changed) Then
			Count = UBound(Changed, 1)
			On Error Resume Next
				For ii = 0 To Count
					Lights(Changed(ii, LAMPNO)).State = Changed(ii, LAMPSTATE)
				Next
				LampCallback
			On Error Goto 0
		End If
	End If
	If UseSolenoids Then
		Changed = Controller.ChangedSolenoids
		If Not IsEmpty(Changed) Then
			Count = UBound(Changed, 1)
			For ii = 0 To Count
				funcName = SolCallback(Changed(ii, SOLNO))
				If funcName <> "" Then Execute funcName & " CBool(" & Changed(ii, SOLSTATE) &")"
			Next
		End If
				On Error Resume Next
						MotorCallback
				On Error Goto 0
	End If
	If isObject(GICallback) Then
		Changed = Controller.ChangedGIStrings
		If Not IsEmpty(Changed) Then
			Count = UBound(Changed, 1)
			For ii = 0 To Count
				GICallback Changed(ii, GISTRINGNO), CBool(Changed(ii, GISTRINGSTATE))
			Next
		End If
	End If
	If UseSync Then Controller.Pause = False
	Me.Enabled = True
End Sub

'----------------------------
' Generic solenoid handlers
'----------------------------
' ----- Flippers ------
Sub vpmSolFlipper(flipObj1, flipObj2, enabled)
	Dim OldSpeed
	If enabled Then
		PlaySound SFlipperOn : flipObj1.RotateToEnd : If Not flipObj2 Is Nothing Then flipObj2.RotateToEnd
	Else
		PlaySound SFlipperOff
		OldSpeed = flipObj1.Speed : flipObj1.Speed = 0.03
		flipObj1.RotateToStart : flipObj1.Speed = OldSpeed
		If Not flipObj2 Is Nothing Then
			OldSpeed = flipObj2.Speed : flipObj2.Speed = 0.03
			flipObj2.RotateToStart : flipObj2.Speed = OldSpeed
		End If
	End If
End sub

' ------ Diverters ------
Sub vpmSolDiverter(divObj, sound, enabled)
	If enabled Then
		divObj.RotateToEnd : If sound Then PlaySound SSolenoidOn
	Else
		divObj.RotateToStart : If sound Then PlaySound SSolenoidOff
	End If
End sub

' ------ Diverters ------
Sub vpmSolWall(wallObj, sound, enabled)
    If IsArray(wallObj) Then
        Dim obj
        For Each obj In wallObj : obj.IsDropped = enabled : Next
    Else
        wallObj.IsDropped = enabled
    End If
    If sound Then If enabled Then PlaySound SSolenoidOn Else PlaySound SSolenoidOff
End sub

' ------- Autoplunger ------
Sub vpmSolAutoPlunger(plungObj, var, enabled)
	Dim oldFire
	If enabled Then
		oldFire = plungObj.FireSpeed : plungObj.FireSpeed = oldFire * (100-var*(2*Rnd-1))/100
		PlaySound SSolenoidOn : plungObj.Fire : plungObj.FireSpeed = oldFire
	Else
		plungObj.Pullback
	End If
End Sub

' ------ Sound Only -------
Sub vpmSolSound(sound, enabled)
	If enabled Then PlaySound sound
End Sub

' ------- Flashers --------
Sub vpmFlasher(flashObj, enabled)
    If IsArray(flashObj) Then
        Dim obj
        For Each obj In flashObj : obj.State = Abs(enabled) : Next
    Else
        flashObj.State = Abs(enabled)
    End If
End Sub

' ------ Magnets ---------
Sub vpmSolMagnet(magnetObj, enabled)
	magnetObj.Enabled = enabled
	If Not enabled Then magnetObj.Kick 180,1
End sub
'-------------------------
' Stubs to allow older games to still work
' These will be removed one day
'
Sub SolFlipper(f1,f2,e) : vpmSolFlipper f1,f2,e : End Sub
Sub SolDiverter(d,s,e) : vpmSolDiverter d,s,e : End Sub
Sub SolSound(s,e) : vpmSolSound s,e : End Sub
Sub Flasher(f,e) : vpmFlasher f,e : End Sub
Sub SolMagnet(m,e) : vpmSolMagnet m,e : End Sub
Function KeyDownHandler(ByVal k) : KeyDownHandler = vpmKeyDown(k) : End Function
Function KeyUpHandler(ByVal k) : KeyUpHandler = vpmKeyUp(k) : End Function

'-- Flipper solenoids (all games)
Const sLRFlipper = 46
Const sLLFlipper = 48
Const sURFlipper = 34
Const sULFlipper = 36

'------------------------------------
' Convert keycode to readable string
'------------------------------------
Dim keyNames
keyNames = array("Escape","1","2","3","4","5","6","7","8","9","0","Minus '-'",_
"Equals '='","Backspace","Tab","Q","W","E","R","T","Y","U","I","O","P","[","]",_
"Enter","Left Ctrl","A","S","D","F","G","H","J","K","L",";","'","`","Left Shift",_
"\","Z","X","C","V","B","N","M",",",".","/","Right Shift","*","Left Menu","Space",_
"Caps Lock","F1","F2","F3","F4","F5","F6","F7","F8","F9","F10","NumLock","ScrlLock",_
"Numpad 7","Numpad 8","Numpad 9","Numpad -","Numpad 4","Numpad 5","Numpad 6",_
"Numpad +","Numpad 1","Numpad 2","Numpad 3","Numpad 0","Numpad .","?","?","?",_
"F11","F12","F13","F14","F15")

Function KeyName(ByVal nKeycode)
	Select Case nKeyCode
		Case 184  KeyName = "R.Alt"
		Case 197  KeyName = "Pause"
		Case 199  KeyName = "Home"
		Case 200  KeyName = "Up"
		Case 201  KeyName = "PageUp"
		Case 203  KeyName = "Left"
		Case 205  KeyName = "Right"
		Case 207  KeyName = "End"
		Case 208  KeyName = "Down"
		Case 209  KeyName = "PageDown"
		Case 210  KeyName = "Insert"
		Case 211  KeyName = "Delete"
		Case Else
			If nKeyCode-1 <= UBound(keyNames) Then KeyName = keyNames(nKeyCode-1) Else KeyName = "?"
	End Select
End Function
'-- Cabinet switches (WPC)
Const swCoin1  = 1
Const swCoin2  = 2
Const swCoin3  = 3
Const swCoin4  = 4
Const swCancel = 5
Const swDown   = 6
Const swUp	   = 7
Const swEnter  = 8
Const swLRFlip = 112
Const swLLFlip = 114
Const swURFlip = 116
Const swULFlip = 118
'-- Key definitions (WPC)
Private Const keyCancel = 8        '(7)    Coin Door Cancel Button
Private Const keyDown = 9          '(8)    Coin Door Down Button
Private Const keyUp = 10           '(9)    Coin Door Up Button
Private Const keyEnter = 11        '(10)   Coin Door Enter Button
Private Const keyInsertCoin1 = 4   '(3)    Inserting a Coin
Private Const keyInsertCoin2 = 5   '(4)    Inserting a Coin
Private Const keyInsertCoin3 = 6   '(5)    Inserting a Coin
Private Const keyInsertCoin4 = 7   '(6)    Inserting a Coin
Private Const keyCoinDoor = 207    '(End)  Opening/Closing Coin Door
Private Const keyStartButton = 2   '(1)    Start Button
Private Const keySlamDoorHit = 199 '(Home) Slam Tilt
Private Const keyShowOpts = 59     '(F1)   Shwoing the Options for Visual PinMAME
Private Const keyShowKeys = 60     '(F2)   Showing Keys
Private Const KeyReset = 61        '(F3)   Reset Emulation
Private Const KeyFrame = 62        '(F4)   Toggle locked window

'-- Keyboard help window (WPC)
Private Sub ShowHelp
	Dim szKeyMsg
	szKeyMsg = "The following keys are defined: " & vbNewLine &_
		vbNewLine & "Visual PinMAME Specific:"  & vbNewLine &_
		KeyName(keyShowOpts) & vbTab & "Game Options" & vbNewLine &_
		KeyName(keyShowKeys) & vbTab & "Key List" & vbNewLine &_
				KeyName(keyReset) & vbTab & "Reset Emulation" & vbNewLine &_
				KeyName(keyFrame) & vbTab & "Toggle window border" & vbNewLine &_
		KeyName(keyStartButton) & vbTab & "Start Button" & vbNewLine &_
		KeyName(keyInsertCoin1)  & vbTab & "Insert Coin #1" & vbNewLine &_
		KeyName(keyInsertCoin2)  & vbTab & "Insert Coin #2" & vbNewLine &_
		KeyName(keyInsertCoin3)  & vbTab & "Insert Coin #3" & vbNewLine &_
		KeyName(keyInsertCoin4)  & vbTab & "Insert Coin #4" & vbNewLine &_
		KeyName(keyCancel) & vbTab & "Coin Door (Escape)" & vbNewLine &_
		KeyName(keyDown)  & vbTab & "Coin Door (Down)" & vbNewLine &_
		KeyName(keyUp)  & vbTab & "Coin Door (Up)" & vbNewLine &_
		KeyName(keyEnter)  & vbTab & "Coin Door (Enter)" & vbNewLine &_
		KeyName(keySlamDoorHit) & vbTab & "Slam Tilt" & vbNewLine &_
		KeyName(keyCoinDoor) & vbTab & "Open/Close Coin Door" & vbNewLine &_
		vbNewLine & "Game Specific" & vbNewLine & ExtraKeyHelp & vbNewLine &_
		vbNewLine & "VP Specific:"  & vbNewLine &_
		KeyName(LeftFlipperKey) & vbTab & "L Flipper" & vbNewLine &_
		KeyName(RightFlipperKey) & vbTab & "R Flipper" & vbNewLine &_
		KeyName(PlungerKey) & vbTab & "Launch Ball" & vbNewLine &_
		KeyName(LeftTiltKey) & vbTab & "Tilt Left" & vbNewLine &_
		KeyName(RightTiltKey) & vbTab & "Tilt Right" & vbNewLine &_
		KeyName(CenterTiltKey) & vbTab & "Tilt Center" & vbNewLine
	MsgBox szKeyMsg,0,"Keyboard Settings..."
End Sub
'-- Keyboard handler (WPC)
Function vpmKeyDown(ByVal keycode)
	' Old games might not define the coin sound
	Dim CoinSound : CoinSound = "" : On Error Resume Next : CoinSound = SCoin : On Error Goto 0
	vpmKeyDown = True ' assume we handle the key
	Select Case keycode
		Case LeftFlipperKey     Controller.Switch(swLLFlip) = True : Controller.Switch(swULFlip) = True
		Case RightFlipperKey    Controller.Switch(swLRFlip) = True : Controller.Switch(swURFlip) = True
		Case keyCancel          Controller.Switch(swCancel) = True
		Case keyDown            Controller.Switch(swDown) = True
		Case keyUp              Controller.Switch(swUp) = True
		Case keyEnter           Controller.Switch(swEnter) = True
		Case keyInsertCoin1     PlaySound CoinSound : Controller.Switch(swCoin1) = True
		Case keyInsertCoin2     PlaySound CoinSound : Controller.Switch(swCoin2) = True
		Case keyInsertCoin3     PlaySound CoinSound : Controller.Switch(swCoin3) = True
		Case keyInsertCoin4     PlaySound CoinSound : Controller.Switch(swCoin4) = True
		Case keyStartButton     Controller.Switch(swStartButton) = True
		Case keySlamDoorHit     Controller.Switch(swSlamTilt) = True
		Case keyCoinDoor        Controller.Switch(swCoinDoor) = Not Controller.Switch(swCoinDoor)
		Case LeftTiltKey        vpmNudge.DoNudge 90,2
		Case RightTiltKey       vpmNudge.DoNudge 270,2
		Case CenterTiltKey      vpmNudge.DoNudge 0,2
		Case Else               vpmKeyDown = False
	End Select
End Function

Function vpmKeyUp(ByVal keycode)
	vpmKeyUp = True ' assume we handle the key
	Select Case keycode
		Case LeftFlipperKey     Controller.Switch(swLLFlip) = False : Controller.Switch(swULFlip) = False
		Case RightFlipperKey    Controller.Switch(swLRFlip) = False : Controller.Switch(swURFlip) = False
		Case keyCancel          Controller.Switch(swCancel) = False
		Case keyDown            Controller.Switch(swDown) = False
		Case keyUp              Controller.Switch(swUp) = False
		Case keyEnter           Controller.Switch(swEnter) = False
		Case keyInsertCoin1     Controller.Switch(swCoin1) = False
		Case keyInsertCoin2     Controller.Switch(swCoin2) = False
		Case keyInsertCoin3     Controller.Switch(swCoin3) = False
		Case keyInsertCoin4     Controller.Switch(swCoin4) = False
		Case keyStartButton     Controller.Switch(swStartButton) = False
		Case keySlamDoorHit     Controller.Switch(swSlamTilt) = False
		Case keyShowOpts
			Controller.Pause = True
			Controller.ShowOptsDialog GetPlayerHWnd
			Controller.Pause = False
		Case keyShowKeys
			Controller.Pause = True
			ShowHelp
			Controller.Pause = False
		Case keyReset
			Controller.Stop : Controller.Run : vpmTimer.Reset
		Case keyFrame
			Controller.LockDisplay = Not CBool(Controller.LockDisplay)
		Case Else	            vpmKeyUp = False
	End Select
End Function

Set vpmTimer = New cvpmTimer
Set vpmNudge = New cvpmNudge


